#-*- coding: utf-8 -*-

"""
    ballroom.decorations
    ~~~~~~~~~~~~~~~~~~~~
    A collection of route decorators

    :authors Mek
    :license pending
"""

import os
import random
import time
import json
from copy import copy
from lazydb import Db
from waltz import web, session, log

def track(fn):
    """A decorator which wraps each route with analytics tracking."""
    def tracked(fn):
        """This setup w/ second inner function allows support for
        passing @track optional arguments and parameters, as well as
        allows for the special case of class methods which require
        self - methods(self).
        """        
        def inner(*args, **kwargs):
            """Copy web context environment, clean it to avoid
            pickeling issues, and dump it to lazydb before returning
            the route
            """
            ctx = copy(web.ctx['env'])
            del ctx['wsgi.errors']
            del ctx['wsgi.input']
            try:
                Db(web.ctx['waltz']['db']).append('analytics', ctx)
            except cPickle.UnpicklingError as e:
                log("Analytics tracking error", lbl="error")
                raise Exception("Database Connection Error: "\
                                    "web.ctx incorrecly configured.")
            return fn(*args, **kwargs)        
        return inner
    return tracked(fn)

def exponential_backoff(exception, err=None, tries=5, debug=False):
    """Exponentially backoff on a certain Exception exception.
    params:
        tries - num of expn backed off attempts before quitting
        err - custom error msg to display in addition to 'e'
              q: are there any reasons why we wouldn't just want
                 to display 'e'? Securiy reasons seem likely.

    note:
        * debug flag used to avoid raising security
          sensitive exception err msgs     
        * sleep may be unsafe (hence adding a psuedo random
          scalar, still bad since not really random)
        * Consider logging errors via
          http://pypi.python.org/pypi/sentry

    usage:
    >>> @exponential_backoff(SomeNetworkError, tries=2)
    ... def foo(bar):
    ...     transfer_file(bar)
    SomeNetworkError: [ExponentialBackoff] Timed out
    after 2 attempts. Network failed to connect. (details: None)
    """
    def decorator(func):
        def inner(*args, **kwargs):
            for n in range(0, tries):
                try:
                    return func(*args, **kwargs)
                except exception as e:
                    scalar = random.randint(0, 1000) / 1000.0
                    time.sleep((2 ** n) + scalar)
            e = e if debug else ''
            raise exception('[Exponential Backoff] ' \
                                'Timed out after %s attempts. ' \
                                '%s (details: %s)' % (tries, e, err))
        return inner
    return decorator

def require_loggedin(web, redir, key='logged', value=True):
    """Func wrapper which returns a decorator for redirecting user to
    appropriate login page if their session[key] != value
    """
    def decorator(f):
        def inner(*args, **kwargs):
            logged = False
            if value in [None, True, False]:
                if session()[key] is value:
                    logged = True
            if session()[key] != value:
                if session()[key] == value:
                    logged = True
            if logged:
                raise web.seeother(redir)
        return inner
    return decorator

def API(web):
    def decorator(f):
        def inner(*args, **kwargs):
            web.header('Content-Type', 'application/json')
            return json.dumps(f(*args, **kwargs))
        return inner
    return decorator



#-*- coding: utf-8 -*-

"""
    waltz.test
    ~~~~~~~~~~

    Test cases for the waltz web framework. Test cases follow webpy's
    test documentation documented here:
    http://webpy.org/cookbook/testing_with_paste_and_nose    
"""

import unittest
import os
import shutil
from paste.fixture import TestApp
import waltz
from waltz import Account, User, web, storage

USERNAME = "username"
PASSWD = "password"
UHASH = 'fe9dfc91b3a89c563a15c1f9d7a1467c08fcb6621a14cfa791014f45bcfac0e3'
SALT = 'mh3ot3si9anq'
EMAIL = "waltz@example.org"
USER_FIELDS = {'age': 24}

_tmpdir = '%s/tmp' % os.path.dirname(__file__)

class TestWaltz(unittest.TestCase):
    
    def setUp(self):
        self.assertTrue(not os.path.exists(_tmpdir))
        os.makedirs(_tmpdir)        
        waltz.setup.init_scaffolding(_tmpdir)
        with open('%s/__init__.py' % _tmpdir, 'w') as f:
            f.write("")

    def test_accounts(self):
        user = Account.register(USERNAME, PASSWD, salt=SALT, **USER_FIELDS)
        self.assertTrue(all(item in user.items() for item in USER_FIELDS.items()),
                        "USER_FIELDS: %s did not persist to user: %s "\
                            % (USER_FIELDS, user))
        self.assertTrue(user.uhash == UHASH,
                        "waltz.Account.register(%s, %s, salt=%s) " \
                            "expected result hash of " \
                            "%s but generated uhash: %s " \
                            % (USERNAME, PASSWD, SALT, UHASH, user.uhash))
        self.assertTrue(user.uhash == UHASH,
                        "waltz.Account.register(%s, %s, salt=%s) " \
                            "expected result  of " \
                            "%s but generated uhash: %s " \
                            % (USERNAME, PASSWD, SALT, UHASH, user.uhash))

    def test_run(self):
        middleware = []
        #waltzapp = __import__() # XXX! '%s/main.py' % _tmpdir
        #self.assertTrue(hasattr(waltzapp, 'env'), "Failed to load waltz app")
        #print waltzapp.env
        #app = TestApp(.wsgifunc(*middleware))
        #r = testApp.get('/')
        #assert_equal(r.status, 200)
        #r.mustcontain('Hello, world!')

    def test_users(self):
        """Test whether waltz.User (of base type Account) interfaces
        with LazyDB correctly and behaves expectedly
        """
        # Setting up waltz web.py env variables for db()
        web.ctx.waltz = storage()
        web.ctx.waltz.db = 'db'

        u1 = User.register(USERNAME, PASSWD, **USER_FIELDS)
        u1.email = EMAIL
        u1.save()
        u2 = User.get(USERNAME)
        u3 = User.register(USERNAME[::-1], PASSWD, **USER_FIELDS)

        self.assertTrue(u1['username'] == USERNAME,
                        "<waltz.User.register> Registration " \
                            "returned invalid user dict")
        print User.get(USERNAME)
        self.assertTrue(User.get(USERNAME).email == EMAIL,
                        "Failed to save / update User record")
        self.assertTrue(u1.authenticate(PASSWD), "User.authenticate Failed")
        self.assertTrue(u2 is not None, "<waltz.User.register> " \
                            "Registration failed, no such entry " \
                            "indexed by {'username': '%s'}" % USERNAME)
        self.assertTrue(all(item in u2.items() for item in USER_FIELDS.items()),
                        "<waltz.User.get> Expected stored data did not match " \
                            "actual data indexed by waltz 'user' LazyDB.")
        User.replace(USERNAME, u3)
        self.assertTrue(User.get(USERNAME)['username'] == USERNAME[::-1],
                        "<waltz.User.replace> Failed to replace username")
        u4 = User.get(USERNAME[::-1], safe=True)
        self.assertTrue(not getattr(u4, 'uhash', False),
                        "User.get(%s, safe=True) did not remove uhash field " \
                            "as would be expected" % USERNAME[::-1])
        users = User.getall()
        self.assertTrue(len(users) == 2, "Users db has incorrect # users: " \
                        "%s (%s)" % (len(users), users))
        users2 = User.getall(safe=True)
        self.assertTrue('uhash' not in users2[USERNAME].keys(),
                        "User.getall(safe=True) did not remove uhash field " \
                            "as would be expected")
        self.assertTrue(User.delete(USERNAME) is not None,
                        "<waltz.User.delete> " \
                            "Failed to delete user: %s" % USERNAME)
    def tearDown(self):
        if os.path.isfile('db'): os.remove('db')
        self.assertTrue(os.path.exists(_tmpdir),
                        "Can't find the tmp directory <Dir: %s>" \
                            ", it should exist!" % _tmpdir)
        shutil.rmtree(_tmpdir)
        self.assertTrue(not os.path.exists(_tmpdir),
                        "tmp directory <Dir: %s> should have been cleaned up" \
                            "/ removed however it still exists!" % _tmpdir)



#-*- coding: utf-8 -*-

"""
    security
    ~~~~~~~~
    Generic User class for interfacing with use accounts +
    authentication.
"""

import hashlib
import random
from utils import Storage, to36, valid_email, \
    ALPHANUMERICS as ALPHAS

username_regex = r'[A-Za-z0-9_%s]{%s,}'
passwd_regex = r'[A-Za-z0-9%s]{%s,}'

class Account(object):
    """The base Account class provides the basic functions for allowing
    user account creation and authentication, however it should be
    extended to allow user retrieval.
    """

    @classmethod
    def authenticate(cls, username, passwd, salt, uhash):
        """Authenticates/validates a user's credentials by comparing
        their username and password versus their computed salted hash.
        A successful authentication results in True, False otherwise.
        """
        return cls._roast(username + salt + passwd) == uhash

    @classmethod
    def register(cls, username, passwd, passwd2=None, salt='', email='', **kwargs):
        """Creates a complete user tuple according to seed
        credentials. The input undergoes a salting and roasting during
        a hashing process and all values are returned for further db
        handling (db insertion, etc). Note: For security reasons,
        plaintext passwords are never returned and should not be
        stored in a db unless there's a very specific reason to.

        XXX: Consider adding the following keyword arguments:
        password_validator - lambda for validating passwd (chars, len)
        username_validator - lambda for validating username (len, etc)

        :param salt: used for testing/verifying hash integrity in tests
        :param kwargs: any addt'l info which should be saved w/ the created user

        usage:
        >>> from waltz import Account
        # typically, salt is not provided as an argument and is instead
        # generated as a side effect of Account.register("username", "password").
        # A salt has been included along with the following function call to
        # guarantee idempotence (i.e. a consistent/same uhash with each call)
        >>> Account.register("username", "password", salt="123456789")
        <Storage {'username': 'username', 'uhash': '021d98a32375ed850f459fe484c3ab2e352fc2801ef13eae274103befc9d0274', 'salt': '123456789', 'email': ''}>
        # using additional kwargs, such as age (i.e. age=24) to add user attributes
        >>> Account.register("username", "password", salt="123456789", age=24)
        <Storage {'username': 'username', 'uhash': '021d98a32375ed850f459fe484c3ab2e352fc2801ef13eae274103befc9d0274', 'salt': '123456789', 'email': '', 'age': 24}>
        """        
        if not passwd: raise ValueError('Password Required')

        if email and not valid_email(email):
            raise ValueError("Email '%s' is malformed and does not " \
                                 "pass rfc3696." % email)
        if passwd2 and not passwd == passwd2:
            raise ValueError('Passwords do not match')
        salt = salt or cls._salt()
        uhash = cls._roast(username + salt + passwd)        
        return Storage(zip(('username', 'salt', 'uhash', 'email'), 
                        (username, salt, uhash, email)) + kwargs.items())

    @classmethod
    def public_key(cls, uid, username, salt):
        """Generates a public key which can be used as a public unique
        identifier token in account activation emails and other
        account specific situations where you wish to create a url
        intended to only work for a specific user.
        """
        return cls._roast(uid + username + salt)

    @classmethod
    def _salt(cls, length=12):
        """http://en.wikipedia.org/wiki/Salt_(cryptography)
        Salting results in the generation of random padding of a
        specified 'length' (12 default) which can be prepended or
        appended to a password prior to hashing to increase security
        and prevent against various brute force attacks, such as
        rainbow-table lookups."""
        return ''.join([random.choice(ALPHAS) for i in range(length)])

    @classmethod
    def _roast(cls, beans, chash=hashlib.sha256):
        """Computes a hash digest from username, salt, and
        password. Hot swappable algo in case there are code changes
        down the road."""
        return chash(beans).hexdigest()




#!/usr/bin/env python

"""
   ballroom.treasury
   ~~~~~~~~~~~~~~~~~

   The treasury contains several modules which, when joined, provide shopping cart function.

   Product:
   A Product is a generic interface which one may extend or inherit in order to
   represent a physical or digital e-commerce product.

   Item:
   A wrapper which encapsulates and functions upon the base Product
   class's attributes in order to interface as a unit of the Cart object.
   Items can be bound to a coupon which may modify or adjust the total
   item price according to the Item.qty (quantity), as well as the 
   specifications of the coupon.

   Cart:
   A shopping Cart is a collection of Items.

       Coupon Calculation:
       The cart has a property() called 'total' which caclulate the
       cart total, with coupons applied, in realtime. First, the
       procedure iterates over each Item within the cart and invoke
       the Item's total property (thus getting the adjusted price for
       the Item, according to the item qty and coupons). After the
       subtotal has been calculated for all Items, any cart level
       coupons are then applied to this subtotal in order to yield the
       final total. Taxes, fees, and commissions are applies at this
       point.

   Coupon:
   Coupons can be applied at both the Cart and the Item level. If
   applied at the Cart level, the coupon will have the effect of a
   value or percentage off of the entire Cart.value. If a Coupon is
   applied to an Item, a 'limit' parameter may be specified as a
   modifier to limit the max number of applications the Coupon can be
   applied to an Item's products (e.g. given an Item having quantity
   5, (i.e. Item.qty == 5) bound to a Coupon having a limit=3 may/can
   only adjust (apply to) 3 of the 5 instances of the Product this
   Item represents.

"""

__author__  = ["Michael E. Karpeles", "Stephen Balaban"]
__email__   = ["michael.karpeles@gmail.com", "mail@stephenbalaban.com"]
__version__ = "0.0.3"

from decimal import Decimal 
from utils import Storage

class Coupon(object):
    def __init__(self, cid, code, percent_off=0, value_off="0.00", pids=None, limit=1):
        """
        params:
            value_off - either a string or a Decimal dollar value with precision of 2 digits
            pids - a list of product ids for which the coupon applies.
                   The default action is for the coupon to apply to the entire 
                   cart / purchase.
            limit - the number of times this coupon should be applied to a product
        """
        self.id = cid # database identifier of the coupon
        self.pids = pids
        self.code = code # coupon code which the user supplies
        self.percent_off = percent_off
        self.value_off = Decimal(value_off)
        self.applications = limit

    @classmethod
    def apply(cls, coupon, price):
        """Applies a coupon to a single unit of an item (an item can
        have a qty > 1) and returns the adjusted price for that
        unit. In terms of order of operations, first subtracts any
        monetary value off discount and then afterwards applies the
        percentage off (if any specified in coupon) to the updated
        subtotal
        """
        if not coupon:
            return price
        dprice = Decimal(.01 * (100.0 - coupon.percent_off)) * (price - coupon.value_off)
        qdprice = (dprice).quantize(Decimal('.01'), rounding='ROUND_DOWN')
        if qdprice <= 0:
            return Decimal('0.00')

class Product(object):
    """Necessary scaffolding for a Product in order to interface with
    the Cart. Feel free to extend this object to fit your needs

    params:
        upc (optional) - Standard UPC-A (see encoding - http://en.wikipedia.org/wiki/Universal_Product_Code#Encoding)
    """
    def __init__(self, pid, label, desc="", price="9.99", currency="USD", phash=None, upc=None):
        self.id = pid
        self.name = label
        self.description = desc
        self.price = Decimal(price)
        self.hash = phash # still needs to be standardized (perhaps see web.py utils for hash ideas)
        self.upc = upc

class Item(object):
    """
    usage:
        >>> p = Product(1, 'shoes', desc='new pair of shoes',
        ...             price='13.37', phash='d81hda8z')
        >>> item = Item(product=p)
        >>> item.total
        '13.37'
    """
    def _verify(self, product=None, qty=1, coupon=None, **kwags):
        """Throws an exception if any of the parameters fail to pass
        verification. A failure may be the result of several
        occurrences, including unexpected types, undesired value
        ranges, signs (i.e. signed versus unsigned), etc.
        """
        #note refid is a user's id and default, -1, is nobody
        if product and not isinstance(product, Product):
            raise TypeError('product argument must inherit from the ' \
                                'Product class.')
        if coupon and not isinstance(coupon, Coupon):
            raise TypeError("Coupon was expected, instead encountered:" \
                                "<type: '%s'>" % type(coupon))
        if type(qty) is not int:
            qty = 1

    def __init__(self, product=None, qty=1, coupon=None, ref=None):
        """
        params:
            product - an instance of an object which subclasses Product
            ref - referral id, for affiliates or referrers
        """
        # need to direclty set to avoid infinite loop
        self._verify(product, qty, coupon)
        object.__setattr__(self, 'qty', qty)
        object.__setattr__(self, 'price', Decimal('0.00'))

        if not ref:
            ref = self.get_empty_ref()
        self.product = product
        self.coupon = coupon
        self.id = product.id
        self.qty = qty
        self.price = getattr(self.product, "price")
        self.currency = getattr(self.product, "currency", "USD")
        self.ref = ref

    def __repr__(self):
        result = "<Item id: "+ str(self.id) + ", total: " + str(self.total) \
                + ", qty: " + str(self.qty) \
                + ", price: " + str(self.price) \
                + ", currency: " + str(self.currency) \
                + ", ref: " + str(self.ref) \
                + ", coupon: " + str(self.coupon) \
                + ", products: " + str(self.product) + " >"
        return unicode(result)

    def __setattr__(self, name, value):
        kwarg = {name: value}
        self._verify(**kwarg)
        object.__setattr__(self, name, value)

    @property
    def total(self):
        ttl = Decimal('0.00')
        applications = self.coupon.applications if self.coupon else 0
        if applications >= self.qty:            
            return Coupon.apply(self.coupon, self.price) * self.qty
        return (Coupon.apply(self.coupon, self.price) * applications) + \
               (self.price * (self.qty - applications))
                
    @classmethod
    def get_empty_ref(self):
        """
        returns an empty ref, free to use!

        >>> Item.get_empty_bag()
        """
        default_refid = -1
        empty_ref = Storage({'id': default_refid,
                             'username': None,
                             'time': None,
                             'uri': None,
                             'tag': None })
        return empty_ref

class Cart(object):
    """
    main atributes:
        total       - total value of cart denominated in cart.currency
        qty         - total number of items in cart
        currency    - cart currency symbol-code tuple ("$", "USD")
        taxrate     - current taxrate of cart (regional basis)
        itemsdict   - list containing current items
    main methods: 
        add(product)
        remove(pid)
        empty()
    usage:
        >>> cart = Cart(itemsdict=[], 
        >>>           taxrate=0, 
        >>>           currency=('$','USD'))
        >>> cart.add(product) # adds a product object to the bar + recalculates
        >>> cart.empty()  # empties cart, resets totals
        >>> cart.total # returns the total cost of all items in cart + tax
        >>> cart.qty # recalculates and returns qty of items in cart
        >>> cart.itemsdict[pid] #returns item object <Item qty: 1, total: blah, product...
        >>> cart.__dict__ # dict representation
        >>> cart.get(product_object) # blah
    
    """
    def _verify(self):
        pass

    def __init__(self, taxrate=Decimal('0.00'), currency=("$", "USD"),
                 qty=0, total=Decimal('0.00'), coupon=None):
        if taxrate > Decimal('1.00') or taxrate < Decimal('0.00'):
            raise ValueError('taxrate is a decimal between 0.00 and 1.00')
        self.taxrate = taxrate
        self.currency = currency
        self.itemsdict = {}
        self.coupon = coupon

    def __repr__(self):
        result = '<Cart total: ' + str(self.total) \
                + ', currency: ' + str(self.currency) \
                + ', itemsdict: ' + str(map(self._repr_wrap, self.itemsdict.keys())) \
                + ', qty: ' + str(self.qty) \
                + ', taxrate: ' + str(self.taxrate) \
                + ', coupon: ' + (str(self.coupon.id) if getattr(self, 'coupon', None) else 'None') \
                + ', pythonid: ' + str(id(self)) + '>'
        return unicode(result)

    def _repr_wrap(self, thing):
        return "<Item " + str(thing) + ">"
    
    def __setattr__(self, name, value):
        """
        controls attribute settings
        """
        if name == 'qty':
            try:
                value = int(value)
            except ValueError as details: 
                raise ValueError("please provide an int for qty")
            if value < 0:
                raise ValueError("qty cannot be less than zero")

        object.__setattr__(self, name, value)

    @property
    def currency_code(self):
        """returns currency code
            >>> cart.currency_code
            "USD"
        """
        return self.currency[1]

    @property
    def currency_symbol(self):
        """returns currency symbol
            >>> cart.currency_symbol
            "$"
        """
        return self.currency[0]

    def add(self, product=None, qty=1, ref=None):
        """ 
        adds product object, sets quantity and refid to cart

        plain vanilla product add
            >>> cart.add(product) 

        add with kwargs:
            >>> cart.add(product=product, qty=3, ref=<Storage ...>)

        """
        if not product: return self
        if self.itemsdict.has_key(product.id):
            #if item exists, increment
            self.itemsdict[product.id].qty += qty

        else:
            item = Item(product=product, qty=qty, ref=ref)
            self.itemsdict[item.id] = item
        self.invariants()
        return self

    @property
    def total(self):
        """Dynamically calculates an up-to-date total in a piecewise
        fashion by traversing each each item in the cart and
        dynamically calculating the totals of each of the cart's
        items. C"""
        subtotal = sum([item.total for k, item in self.itemsdict.items()])
        return Coupon.apply(self.coupon, subtotal)

    @property
    def qty(self):
        """Returns the number of unique items within the cart ignoring the
        quantity of the items, itself.

        XXX Make sure len is working correctly over list of tuples
        returned by itemsdict.items()
        """
        return sum([item.qty for k, item in self.itemsdict.items()])

    def remove(self, pid, amt=None):
        """removes ALL of an item from cart based on product id 

        >>> #using id
        >>> cart.remove(2) #pid = 2, an item in cart
        2
        >>> cart.remove(2) # an item no longer in cart raises KeyError
        Traceback (most recent call last):
            File "<stdin>", line 1, in <module>
            File "cart/cart.py", line 149, in remove
                raise KeyError('no item found')
            KeyError: 'no item found'

        Note: if you only want to remove some of the items in the cart,
        use the cart.set(id, qty) method instead.
        """
        if type(pid) is not (long or int):
            raise TypeError("Product ID must be an int or long; " \
                                "Instead received value: %s of " \
                                "type %s" % (pid, type(pid)))
        if not self.itemsdict.has_key(pid):
            raise KeyError('no item found')
        else:
            del self.itemsdict[pid]
        self.invariants()
        return self

    def get(self, pid=None, slug=None):
        """
        gets an Item based on product_id
            >>> cart.get(pid) #pid = 2 (an item not in cart)
            Traceback (most recent call last):
                File "<stdin>", line 1, in <module>
                File "cart/cart.py", line 171, in get
                    return self.itemsdict[key]
                KeyError: 2
            >>> cart.get(2) #an item in cart
            <Item id: 2, total: ... >
        """
        pid = long(pid)
        key = -1
        if pid is None and slug is None:
            raise ValueError("please supply an pid or slug")
        if pid is None:
            # slug search O(n)
            for k,v in self.itemsdict.items():
                if v.product.name == slug:
                    key = k
        else:
            # pid O(1)
            key = pid
        return self.itemsdict[key]

    def contains(self, pid=None):
        """
        returns boolean if itemsdict contains product id (pid)

        consider:
           extend this function s.t. you provide a key and a value to
           search for.

        usage:
            >>> cart.add(product) # product.id = 3, product.slug = "Slug Text"
            >>> cart.contains(3)
            True
        """
        if pid:
            raise ValueError("please supply an pid or slug")
        return self.itemsdict.has_key(pid)

    @property
    def tax(self):
        return Decimal(self.total * self.taxrate).quantize(Decimal('.01'), rounding='ROUND_DOWN')

    def invariants(self):
        """tests cart invariants -- TODO"""
        pass

    def empty(self):
        """
        empties this cart, keeps taxrate, currency, returns copy
            >>> cart.empty()
            <Cart total: 0.00, currency: ('$', 'USD'), qty: 0, itemsdict: {},
            taxrate: 0.00, coupon: None>
        """
        return self.nullify()

    def nullify(self):
        """
        Purges all new carts directly after being issued to a session
        """
        self.coupon = None
        self.itemsdict = {} 
    
    def make_receipt(self):
        """ 
        Return a string detailing all of the items in the cart
        and their costs to be used as a receipt.
        """
        desc = ""
        for index, item in enumerate(self.itemsdict):
            if index == 0:
                desc = "item #%s - %s - $%s" % (self.itemsdict[item].product.hash,
                                                self.itemsdict[item].product.name,
                                                self.itemsdict[item].price)
            else:
                desc = "%s, item #%s - %s - $%s" % (desc, self.itemsdict[item].product.hash,
                                                    self.itemsdict[item].product.name,
                                                    self.itemsdict[item].price)
        return desc




#!/usr/bin/env python

"""waltz: make web apps in 3/4 time (http://github.com/mekarpeles/waltz)"""

__version__ = "0.1.7"
__author__ = [
    "Mek <michael.karpeles@gmail.com>"
]
__license__ = "public domain"
__contributors__ = "see AUTHORS"

import logging
import web
from lazydb import Db

def logger(logpath, msg, level=logging.INFO, method='info'):
    logging.basicConfig(filename=logpath, level=level)
    getattr(logging, method, 'info')(msg)

session = lambda: getattr(web.ctx, 'session', None)
# render: renders a template through base template base.html
# (unless base template name is overridden)
render = lambda: getattr(web.ctx, 'render', None)
# slim render: renders a template by itself with no base template
slender = lambda: getattr(web.ctx, 'slender', None)
# db for waltz analytics, etc.
db = lambda: Db(web.ctx['waltz']['db'])
log = lambda msg, lbl='info': logger(web.ctx['waltz']['logging'], msg, method=lbl)

from security import Account
from decorations import *
from utils import *
from treasury import *
from setup import *
from modules import rss
from utils import Storage

class User(Account, Storage):
    """Extends Account to use LazyDB as Datastore"""

    udb = 'users'
    db = staticmethod(db)

    def __init__(self, uid, user=None):
        """
        TODO:
        * Provide ability to self.save()
        * Extend __init__ to populate User obj as Storage
        """
        u = user if user else self.get(uid)
        if u is None:
            raise AttributeError("No user found with id: %s" % uid)
        for k, v in u.items():
            setattr(self, k, v)


    def __repr__(self):     
        return '<User ' + repr(self._publishable(dict(self))) + '>'

    def save(self):
        """save the state of the current user in the db; replace
        existing databased user (if it exists) with this instance of
        the User or insert this User otherwise
        """
        users = User.getall()
        users[self.username] = dict(self)
        return self.db().put(self.udb, users)

    @classmethod
    def getall(cls, safe=False):
        users = cls.db().get(cls.udb, default={}, touch=True)
        for uid, user in users.items():           
            u = (user if not safe else cls._publishable(user))
            users[uid] = cls(uid, user=u)
        return users

    @classmethod
    def get(cls, uid, safe=False):
        """
        params:
            uid - the user id which to fetch
            safe - return users without secure fields like salt and hash
            db - provide your own instantiated lazydb.Db() connector
        """
        if uid is not None:
            users = cls.getall()
            try:
                return users[uid] if not safe else cls._publishable(users[uid])
            except KeyError:
                return None

    @classmethod
    def _publishable(cls, usr, *args):
        """Make this user publishable. Removes fields from a user
        which would be undesirable to publish publicly, like salt or
        hash. Can take an arbitrary list of additional keys to
        scrub/remove"""
        keys = set(args + ('uhash', 'salt'))
        for key in keys:
            if key in usr:
                del usr[key]
        return usr

    @classmethod
    def insert(cls, usr, pkey='username'):
        """Appends usr to users in db and returns the
        id of the new user.
        
        params:
            usr - dictionary or Storage
        """
        users = cls.db().get(cls.udb, default={}, touch=True)
        uid = usr[pkey]
        users[uid] = usr
        users = cls.db().put(cls.udb, users)
        return uid

    @classmethod
    def replace(cls, uid, usr):
        users = cls.db().get(cls.udb, {}, touch=True)
        users[uid] = usr
        cls.db().put(cls.udb, users)
        return usr

    @classmethod
    def update(cls, uid, func=lambda x: x):
        """Updates a given user by applying a func to it. Defaults to
        identity function
        """
        users = cls.db().get(cls.udb, default={}, touch=True)
        user = func(users[uid])
        users[uid] = user
        cls.db().put('users', users)
        return user

    @classmethod
    def delete(cls, uid):
        users = cls.db().get(cls.udb, default={}, touch=True)
        try:
            del users[uid]
            return cls.db().put(cls.udb, users)
        except KeyError:
            return False

    def authenticate(self, passwd):
        """Instance level authentication for a User

        usage:
        >>> u = User('username')
        >>> u.authenticate('password')
        True
        """
        return self.easyauth(dict(self), passwd)

    @classmethod
    def easyauth(cls, u, passwd):
        """New-style auth which takes a user dict and a passwd

        params:
            u - a user dict with (at least) items:
                ['salt', 'uhash', 'username']
        """
        if u and all(key in u for key in ['username', 'salt', 'uhash']):
            return Account.authenticate(u['username'], passwd,
                                        u['salt'], u['uhash'])
        raise TypeError("Account._auth expects user object 'u' with " \
                            "keys: ['salt', 'uhash', 'username']. " \
                            "One or more items missing from user dict u.")

    @classmethod
    def register(cls, username, passwd, passwd2=None, email='', salt='',
                 enabled=True, pkey="username", **kwargs):
        """Calls Account's regiser method and then injects **kwargs
        (additional user information) into the resulting dictionary.
        
        XXX Additional constraints required (like check for multiple
        emails + other keys which should be 'unique')

        :param pkey: the string name of the key which will be used as
                     a primary key for storing/referencing/indexing
                     this user within lazydb
        usage:
        >>> from waltz import User
        >>> User.register("username", "password", passwd2="password",
        ...               email="username@domain.org", salt='123456789',
        ...               pkey="email")
        <Storage {'username': 'username',
        'uhash': '021d98a32375ed850f459fe484c3ab2e352fc2801ef13eae274103befc9d0274', 
        'salt': '123456789', 'email': 'username@domain.org'}>
        """
        #if not cls.registered(username):
        user = super(User, cls).register(username, passwd, passwd2=passwd2,
                                         email=email, salt=salt, **kwargs)
        user.enabled = enabled
        uid = cls.insert(user, pkey=pkey)
        return cls(user.username, user=user)

    @classmethod
    def registered(cls, username):
        """predicate which answers whether a username exists in the User db.

        XXX registered() should be made more flexible to accomomdate
        for keys other than username in the case where there are
        UNIQUEness constraints on user attributes, such as email. This
        may mean providing **kwargs of unique User attributes and
        values and iterating over each user in the Db (this linear
        approach is a caveat of lazydb) to ensure UNIQUEness.

        >>> from waltz import User
        >>> User.registered("username")
        True
        >>> User.registered("Guido van Rossum")
        False
        """
        if any(usr['username'] == username for usr in cls.getall().values()):
            return True
        return False

    @classmethod
    def change_db(cls, dbname):
        """Internally sets the name of the database to use"""
        setattr(cls, 'db', staticmethod(lambda: Db(dbname)))



#-*- coding: utf-8 -*-
"""
    static
    ~~~~~~
    Static html resources to render during scaffolding
"""

index = """$def with()

$# Does this specific template (index.html) require its own css or js dependencies?
$# Uncomment and or modify the following two lines by removing the # symbol

$#css = /static/css/style.css
$#js = /static/js/main.js

<h1>Darling, we're ready to waltz!</h1>
"""

style= """
"""

base = """$def with (content)

<!DOCTYPE html>
<!--[if lt IE 7]> <html class="no-js lt-ie9 lt-ie8 lt-ie7"> <![endif]-->
<!--[if IE 7]> <html class="no-js lt-ie9 lt-ie8"> <![endif]-->
<!--[if IE 8]> <html class="no-js lt-ie9"> <![endif]-->
<!--[if gt IE 8]><!--> <html class="no-js"> <!--<![endif]-->
  <head>
    <title></title>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge,chrome=1">
    <meta name="viewport" content="width=device-width">
    <meta name="description" content="">
    <link rel="stylesheet" href="/static/css/style.css">
    $# Dynamically link js files related to $content
    $for src in content.get('js', "").split(" "):
      $if src:
        <script type="text/javascript" src="$src"></script>
    $# Dynamically link css files related to $content
    $for href in content.get('css', "").split(" "):
      $if href:
        <link rel="stylesheet" href="$href" />
  </head>
  <body>
    <!--[if lt IE 7]>                         
        <p class="chromeframe">You are using an <strong>outdated</strong> browser. Please 
	  <a href="http://browsehappy.com/">upgrade your browser</a> or <a href="http://www.google.com/chromeframe/?redirect=true">
	    activate Google Chrome Frame</a> to improve your experience.</a>
        <![endif]-->

    <header>
    </header>

    <div id="content">
      $:content
    </div>

    <script> $# Google Analytics ====================================
      var _gaq=[['_setAccount','XX-XXXXXXXX-X'],['_trackPageview']];
      (function(d,t){var g=d.createElement(t),s=d.getElementsByTagName(t)[0];
      g.src=('https:'==location.protocol?'//ssl':'//www')+'.google-analytics.com/ga.js';
      s.parentNode.insertBefore(g,s)}(document,'script'));
    </script>
  </body>
</html>
"""

mainpy = """#!/usr/bin/env python
#-*- coding: utf-8 -*-

\"\"\"
    main.py
    ~~~~~~~

    Main waltz application.

    :copyright: (c) __ by __.
    :license: __ , see LICENSE for more details.
\"\"\"

import waltz

urls = ('/analytics/?', 'waltz.modules.Analytics',
        '/?', 'routes.home.Index')

sessions = {}
env = {}
app = waltz.setup.dancefloor(urls, globals(), sessions=sessions, env=env)

if __name__ == "__main__":
    app.run()
"""

homepy = """#-*- coding: utf-8 -*-

\"\"\"
    routes.home
    ~~~~~~~~~~~

    Routes related to the homepage.

    :copyright: (c) __ by __.
    :license: __ , see LICENSE for more details.
\"\"\"

import waltz
from waltz import track, render

class Index:
    @track
    def GET(self):
        return render().index()
"""



"""
   modules
   ~~~~~~~
   Modules/plugins + extras for the waltz framework
"""

from waltz import web, db
from datetime import datetime     
from lazydb import Db

class Analytics:
    def GET(self):
        web.header('Content-Type', 'application/json')
        return db().get('analytics')

def rss(items_func, template=None, **kwargs):
    rss = RSS(template=template, **kwargs)

    class RSSfeed:
        def GET(self):
            web.header('Content-Type', 'application/xml')
            return rss.feed(items_func())
    return RSSfeed

class RSS:

    _template = """$def with (title="", link="", description="", \
 language="en-us", date="", generator="", \
 editor="", webmaster="", items=[])

<!--?xml version="1.0"?-->
<rss version="2.0">
   <channel>
      <title>$title</title>
      <link>$link</link>
      <description>$description</description>
      <language>$language</language>
      <pubdate>$date</pubdate>

      <lastbuilddate>$date</lastbuilddate>
      <generator>$generator</generator>
      <managingeditor>$editor</managingeditor>
      <webmaster>$webmaster</webmaster>

      $for item in items:
        <item>
           <title>$(item['title'] if 'title' in item else '')</title>
           <link>$(item['link'] if  'link' in item else '')</link>
           <description><!--[CDATA[$(item['description'][:100] if 'description' in item else '')..]]--></description>
           <pubdate>$(item['date'] if 'date' in item else '')</pubdate>
           <guid>$(item['guid'] if 'guild' in item else '')</guid>
        </item>
    </channel>
</rss>
"""
    def __init__(self, template=None, **kwargs):
        template = template or self._template
        self.kwargs = kwargs
        if template == self._template:
            for (k, v) in [('title', ''),
                           ('description', ''),
                           ('language', 'en-us'),
                           ('date', datetime.now().ctime()),
                           ('generator', ''),
                           ('editor', 'waltz'),
                           ('webmaster', '')]:
                kwargs.setdefault(k, v)
        self.template = web.template.Template(template)

    def feed(self, items):
        """Example items list for waltz RSS template:

        [{'title': '',
          'link': '',
          'body': '',
          'description': '',
          'date': '',
          'guid': ''
          }, ...
        ]

        :param items: a function which generates item dicts of the
                      above format
        """
        kwargs = {'items': items}
        kwargs.update(self.kwargs)     
        return self.template(**kwargs)



#-*- coding: utf-8 -*-

"""
    ballroom.setup
    ~~~~~~~~~~~~~~
    Setup webpy app with additional waltz;
    Sessions and authentication, templates, etc.
"""

import os
from functools import partial
import web
from web import wsgiserver
from reloader import PeriodicReloader
_https = wsgiserver.CherryPyWSGIServer

def dancefloor(urls, fvars, sessions=False, autoreload=False,
               debug=True, **kwargs):
    """
    params:
        urls - a flat tuple consisting of alternating strings pairs
               representing a mapping between a route's regex match
               and its handling Class ("/route-regex", "ClassHander", ...)
        fvars - globals() containing __file__ or __main__ sent from
                root main.py app
        sessions - a boolean False denotes sessions will not be used.
                   Otherwise, a dict is expected which is used as
                   the default session data structure / values.
    **kwargs:
        env - a dict of environment ctx vars + funcs which will
              be made globally accesible from within html templates
        session_store - can be overridden with a web.session storage
                        method s.t. the user may use DBStore or specify
                        an alternate path besides the default, 'sessions/'
        session - a dictionary representing a default init'd session
    """
    _path = os.path.dirname(os.path.realpath(fvars['__file__']))
    app = web.application(_preprocess(urls), fvars, autoreload=autoreload)
    env = {'ctx': web.ctx}
    env.update(kwargs.get('env', {}))

    if type(kwargs.get('ssl', None)) is tuple:        
        _https.ssl_certificate, _https.ssl_private_key = kwargs.get('ssl')

    def setup_rendering():
        html = partial(web.template.render, '%s/templates/' % _path)
        slender = html(globals=env)
        render = html(base='base', globals=env)
        def render_hook():
            web.ctx.render = render
            web.ctx.slender = slender
        app.add_processor(web.loadhook(render_hook))
        env['render'] = slender

    def setup_sessions():
        if sessions is False:
            env['session'] = None
            return
        def default_store():
            """Default method of storing session: DiskStore
            created directory sessions/ by default to store sessions"""
            path = _path + '/sessions'
            if not os.path.exists(path):
                os.makedirs(path)
            return web.session.DiskStore(path)

        store = kwargs.get('session_store', default_store())
        session = init_sessions(web, app, store, sessions)
        env['session'] = session

    def setup_waltz():
        def fcgi(func, addr=None):
            """nginx with spawn-fcgi"""
            return web.wsgi.runfcgi(func, addr)
        web.config.debug = debug
        if debug:
            PeriodicReloader()
        if kwargs.get('fcgi'):
            web.wsgi.runwsgi = fcgi
        db = kwargs.get('db', "%s/db" % _path)
        lgr = kwargs.get('logging', '%s/events.log' % _path)
        def waltz_hook():
            web.ctx.waltz = {"debug": debug,
                             "db": db,
                             "logging": lgr
                             }
        app.add_processor(web.loadhook(waltz_hook))        

    setup_rendering()
    setup_sessions()
    setup_waltz()
    return app

def init_sessions(web, app, store, session):
    """kwargs is used to inject options like 'cart' into session."""
    web.config.session_parameters['ignore_expiry'] = True
    session = web.session.Session(app, store, initializer=session)
    def inject_session():
        """closure; uncalled function which wraps session is
        passed to the web loadhook and invoked elsewhere and at a
        later point in time
        """
        web.ctx.session = session
    app.add_processor(web.loadhook(inject_session))
    return session

def init_scaffolding(_path, appname="main.py", **kwargs):
    """Builds scaffolding for the project:
    static/, templates/, routes/, subapps/"""

    def build_static():
        """Builds a static/ directory within the project
        with subdirs: static/css, static/js, and static/imgs"""
        from waltz.static import style
        path = _path + '/static'
        if not os.path.exists(path):
            os.makedirs(path)
        for subdir in ['css', 'js', 'imgs']:
            subpath = '%s/%s' % (path, subdir)
            if not os.path.exists(subpath):
                os.makedirs(subpath)
        stylecss = path + "/css/style.css"
        if not os.path.exists(stylecss):
            with open(stylecss, 'w') as f:
                f.write(style)

    def build_routes():
        """Creates directories + __init__ files for route logic"""
        for d in ['routes', 'subapps']:
            path = '%s/%s' % (_path, d)
            if not os.path.exists(path):
                os.makedirs(path)
                fname = '%s/%s' % (path, '__init__.py')
                with open(fname, 'w') as f: f.close()

    def build_templates():
        """Builds the templates/ directory"""
        from waltz.static import base, index
        path = '%s/templates' % _path
        if not os.path.exists(path):
            os.makedirs(path)
        for fname, content in [('base', base), ('index', index)]:
            base = '%s/%s.html' % (path, fname)
            if not os.path.exists(base): 
                with open(base, 'w') as f:
                    f.write(content)
    
    def build_app(appname):
        from waltz.static import mainpy, homepy
        appname += ".py" if appname[-3:] != ".py" else ""
        homepypath = '%s/routes/home.py' % _path
        mainpypath = '%s/%s' % (_path, appname)
        for fname, content in [(mainpypath, mainpy),
                               (homepypath, homepy)]:
            if not os.path.exists(fname): 
                with open(fname, 'w') as f:
                    f.write(content)            

    build_static()
    build_routes()
    build_templates()
    build_app(appname)

def _preprocess(urls):
    """Can be used to inject routes in the future"""
    return urls

